package	jp.co.sds_corp.fixconnect.message;

import	java.util.HashMap;
import	java.util.List;
import	java.util.Map;

import	jp.co.sds_corp.fixconnect.FIXException;
import jp.co.sds_corp.util.config.DOMConfiguration;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class LayoutBuilder
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	Map<String,Config> mapConfig;
	static class Config {
		String	msgType;
		String	name;
		IConfiguration	config;
	}
	IConfiguration	headerConfig;
	IConfiguration	trailerConfig;

	private static LayoutBuilder instance = new LayoutBuilder();
	private	boolean initialized = false;

	private LayoutBuilder() {
	}

	public static LayoutBuilder getInstance() throws FIXException {
		synchronized( instance ) {
			if ( !instance.initialized ) {
				instance.mapConfig = new HashMap<String,Config>();
				try {
					instance.readConfig();
				} catch ( Exception ex ) {
					throw( new FIXException( "config load error.", ex ));
				}
				instance.initialized = true;
			}
		}
		return instance;
	}

	void readConfig() throws Exception {
		DOMConfiguration doc = DOMConfiguration.findResource( "fixmessage.xml" );
		headerConfig = doc.getChild( "FIXHeader" );
		if ( headerConfig == null ) {
			throw( new Exception( "header config not found." ));
		}
		trailerConfig = doc.getChild( "FIXTrailer" );
		if ( trailerConfig == null ) {
			throw( new Exception( "trailer config not found." ));
		}

		List<IConfiguration> list = doc.listChild( "message" );
		for ( int i = 0; i < list.size(); i++ ) {
			IConfiguration element = list.get( i );
			Config msgCfg = new Config();
			msgCfg.msgType = element.getAttributes().get( "MsgType" );
			msgCfg.name = element.getAttributes().get( "name" );
			msgCfg.config = element;
			logger.debug( this, "load message type=" + msgCfg.msgType
				+ " name=" + msgCfg.name );
			mapConfig.put( msgCfg.msgType, msgCfg );
		}
	}

	/**
	* 電文定義を読み込みます。
	*/
	void createLayout( FIXMessage msg, String msgType )
		throws FIXException
	{
		Layout layout = msg.getLayout();
		Config config = mapConfig.get( msgType );
		if ( config == null ) {
			throw( new FIXException( "MessageType[" + msgType + "] "
						+ "not found." ));
		}

		createLayout( layout, headerConfig );
		createLayout( layout, config.config );
		createLayout( layout, trailerConfig );

		msg.setMsgType( msgType );
		msg.setMsgTypeName( config.name );
		msg.setField( "MsgType", msgType );
	}

	void createLayout( Layout layout, IConfiguration config )
		throws FIXException
	{
		try {
			List<IConfiguration> list = config.listChild();
			for ( int i = 0; i < list.size(); i++ ) {
				IConfiguration element = list.get( i );
				Field field = createField( element );
				if ( field != null ) {
					layout.addField( field );
				}
			}
		} catch ( Exception ex ) {
			throw( new FIXException( "create layout error.", ex ));
		}
	}

	protected Field createField( IConfiguration element )
		throws Exception
	{
		String tagname = element.getName();
		if ( tagname.equals( "component" )) {
			Component component = createComponent( element );
			return component;
		} else
		if ( !tagname.equals( "tag" )) {
			logger.warn( this, "unknown tag[" + tagname + "]" );
			return null;
		}
		int id = -1;
		String sid = element.getAttributes().get( "id" );
		String name = element.getAttributes().get( "name" );
		String def = element.getAttributes().get( "default", null );
		try {
			id = Integer.valueOf( sid ).intValue();
		} catch ( Exception ex ) {
			throw( new FIXException( "id(" + sid + ")"
						+ " must be number.", ex ));
		}

		Tag tag = new Tag( id, name );
		Field ret = new Field( tag );
		if ( def != null && def.length() > 0 ) {
			ret.setDefaultValue( def );
		}
		return ret;
	}

	protected Component createComponent( IConfiguration element )
		throws Exception
	{
		Component ret = null;
		String name = element.getAttributes().get( "name" );

		int count = -1;
		String scount = element.getAttributes().get( "count", null );
		if ( scount != null && scount.length() > 0 ) {
			try {
				count = Integer.valueOf( scount ).intValue();
			} catch ( NumberFormatException ex ) {
			}
		}

		if ( count > 0 ) {
			ret = new Component( name, count );
		} else {
			ret = new Component( name );
		}
		try {
			IConfiguration repeat = element.getChild( "repeat" );
			if ( repeat != null ) {
				IConfiguration cntCfg = element.getChild( "tag" );
				Field cnt = createField( cntCfg );
				cnt.setCountField( true );
				cnt.setComponent( ret );
				ret.setCountField( cnt );
				ret.setRepeat( true );

				List<IConfiguration> list = repeat.listChild();
				for ( int i = 0; i < list.size(); i++ ) {
					IConfiguration child = list.get( i );
					Field field = createField( child );
					if ( field != null ) {
						field.setComponent( ret );
						ret.addField( field );
					}
				}
			} else {
				List<IConfiguration> list = element.listChild();
				for ( int i = 0; i < list.size(); i++ ) {
					IConfiguration child = list.get( i );
					Field field = createField( child );
					if ( field != null ) {
						field.setComponent( ret );
						ret.addField( field );
					}
				}
			}
		} catch ( Exception ex ) {
			throw( new FIXException( "create layout error.", ex ));
		}
		return ret;
	}

}
