package	jp.co.sds_corp.fixconnect.transport;

import java.net.Socket;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;
import java.util.Vector;
import java.util.concurrent.atomic.AtomicReference;

import jp.co.sds_corp.fixconnect.FIXException;
import jp.co.sds_corp.fixconnect.logger.IMessageLogger;
import jp.co.sds_corp.fixconnect.message.FIXMessage;
import jp.co.sds_corp.fixconnect.message.IMessageParser;
import jp.co.sds_corp.fixconnect.message.IMessageReceiver;
import jp.co.sds_corp.fixconnect.message.MessageReceiveEvent;
import jp.co.sds_corp.fixconnect.message.SessionRejectReason;
import jp.co.sds_corp.fixconnect.store.IMessageStore;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class FIXSession implements IMessageReceiver, ITransportEventListener
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	IMessageStore	store;
	Vector<IMessageLogger>	vecLogger = new Vector<IMessageLogger>();
	TransportSupport transportSupport = new TransportSupport();
	StateSupport stateSupport = new StateSupport();
	ReceiverSupport recvSupport = new ReceiverSupport();
	AtomicReference<SessionState> refState =
		new AtomicReference<SessionState>( SessionState.IDLE );
	HeartBeatChecker	hbChecker;

	public FIXSession() {
		timeout = 5000;
		this.sessionId = String.valueOf( hashCode() );
	}

	public FIXSession( Socket sock ) {
		this();
		this.socket = new FIXSocket( sock );
	}

	public void init( IConfiguration config ) throws Exception {
		String var;
		var = config.getAttributes().get( "debug" );
		if ( var.equals( "true" )) {
			setDebug( true );
		}

		IConfiguration c;
		c = config.getChild( "Parser" );
		parser = createParser( c );

		c = config.getChild( "Store" );
		store = createStore( c );

		List<IConfiguration> list = config.listChild( "Logger" );
		for ( IConfiguration lc : list ) {
			IMessageLogger logger = createLogger( lc );
			vecLogger.add( logger );
		}

		c = config.getChild( "Session" );
		initSession( c );
	}

	void initSession( IConfiguration config ) throws Exception {
		heartbeat = 30;
		heartbeatTimeout = 10;

		IConfiguration attr = config.getAttributes();
		String var;
		version = attr.get( "version" );
		transportProtocol = attr.get( "transport", null );
		var = attr.get( "heartbeat" );
		if ( var != null && var.length() > 0 ) {
			heartbeat = Long.valueOf( var ).longValue();
		}
		var = attr.get( "heartbeatTimeout", null );
		if ( var != null && var.length() > 0 ) {
			heartbeatTimeout = Long.valueOf( var ).longValue();
		}
		senderCompID = attr.get( "SenderCompID" );
		targetCompID = attr.get( "TargetCompID" );

		hbChecker = new HeartBeatChecker( this );
		addMessageReceiver( hbChecker );
	}

	IMessageParser createParser( IConfiguration config ) throws Exception {
		IMessageParser parser = null;
		try {
			Class cls = Class.forName( config.getAttributes().get( "class" ));
			parser = (IMessageParser )cls.newInstance();
			parser.init( config );
		} catch ( Exception ex ) {
			throw( new FIXException( "parser create error", ex ));
		}
		return parser;
	}

	IMessageStore createStore( IConfiguration config ) throws Exception {
		IMessageStore store = null;
		try {
			Class cls = Class.forName( config.getAttributes().get( "class" ));
			store = (IMessageStore )cls.newInstance();
			store.init( config );
		} catch ( Exception ex ) {
			throw( new FIXException( "store create error", ex ));
		}
		return store;
	}

	IMessageLogger createLogger( IConfiguration config ) throws Exception {
		IMessageLogger logger = null;
		try {
			Class cls = Class.forName( config.getAttributes().get( "class" ));
			logger = (IMessageLogger )cls.newInstance();
			logger.init( config );
		} catch ( Exception ex ) {
			throw( new FIXException( "logger create error", ex ));
		}
		return logger;
	}

	/**
	* セッション状態を設定します。
	* @param	expected	予想される状態
	* @param	var		設定するデータ
	*/
	void setState( SessionState expected, SessionState var ) {
		if ( !refState.compareAndSet( expected, var )) {
			return;
		}
		fireStateChanged( expected, var );
	}
	/**
	* セッション状態を設定します。
	* @param	var		設定するデータ
	*/
	void setState( SessionState var ) {
		fireStateChanged( var );
		refState.set( var );
	}
	/**
	* セッション状態を取得します。
	* @return	セッション状態
	*/
	public SessionState getState() {
		return refState.get();
	}

	public void start() throws FIXException {
		socket.setParser( parser );
		socket.addMessageReceiver( this );
		socket.addEventListener( this );
		socket.start();

		hbChecker.addEventListener( this );
		hbChecker.start();
	}

	public void close() throws FIXException {
		hbChecker.stop();
		socket.close();
		try {
			store.close();
		} catch ( Exception ex ) {
			logger.error( this, "store close error", ex );
		}
		for ( IMessageLogger mlog : vecLogger ) {
			try {
				mlog.close();
			} catch ( Exception ex ) {
				logger.error( this, "logger close error", ex );
			}
		}
		socket.removeMessageReceiver( this );
		socket.removeEventListener( this );
		hbChecker.removeEventListener( this );
	}

	public void onEvent( TransportEvent ev ) {
		fireEventOccurred( ev );
		switch ( ev.getType()) {
		case CONNECT:
			setState( SessionState.CONNECTED );
			break;
		case CLOSE:
			socketClosed( ev.getSocket() );
			setState( SessionState.IDLE );
			break;
		default:
		}
	}

	public void onException( FIXException ex ) {
		fireExceptionOccurred( ex );
	}

	void socketClosed( FIXSocket socket ) {
		logger.info( this, "socket closed. "
			+ "session id=" + getSessionId() + " socket=[" + socket + "]" );
		hbChecker.stop();
		if ( !socket.isClosed()) {
			try {
				socket.close();
			} catch ( Exception ex ) {
				logger.error( this, "socket close error", ex );
			}
		}
	}

	public void addEventListener( ITransportEventListener l ) {
		transportSupport.addEventListener( l );
	}
	public boolean removeEventListener( ITransportEventListener l ) {
		return transportSupport.removeEventListener( l );
	}
	public void fireEventOccurred( TransportEvent ev ) {
		transportSupport.fireEventOccurred( this, ev );
	}
	public void fireExceptionOccurred( FIXException ex ) {
		transportSupport.fireExceptionOccurred( ex );
	}

	public void addStateChangeListener( IStateChangeListener l ) {
		stateSupport.addStateChangeListener( l );
	}
	public boolean removeStateChangeListener( IStateChangeListener l ) {
		return stateSupport.removeStateChangeListener( l );
	}
	public void fireStateChanged( SessionState newState ) {
		stateSupport.fireStateChanged( this, getState(), newState );
	}
	public void fireStateChanged( SessionState oldState, SessionState newState ) {
		stateSupport.fireStateChanged( this, oldState, newState );
	}
	public void fireStateChanged( StateChangeEvent ev ) {
		stateSupport.fireStateChanged( ev );
	}

	public void addMessageReceiver( IMessageReceiver l ) {
		recvSupport.addMessageReceiver( l );
	}
	public boolean removeMessageReceiver( IMessageReceiver l ) {
		return recvSupport.removeMessageReceiver( l );
	}
	public void addMessageReceiver( String msgType, IMessageReceiver l ) {
		recvSupport.addMessageReceiver( msgType, l );
	}
	public boolean removeMessageReceiver( String msgType, IMessageReceiver l ) {
		return recvSupport.removeMessageReceiver( msgType, l );
	}
	public void fireMessageReceived( MessageReceiveEvent ev ) {
		recvSupport.fireMessageReceived( this, ev );
	}

	public void onMessage( MessageReceiveEvent ev ) {
		ev.setSession( this );
		FIXMessage msg = ev.getMessage();
		String msgType = msg.getMsgType();
		try {
			// getSocket().getParser().validate( msg );
			log( IMessageLogger.Direction.RECV, msg );
			String seq = (String )msg.getField( "MsgSeqNum" );
			store.setRecvSeqNum( Integer.valueOf( seq ) );

			if ( msgType.equals( "2" )) {
				resend( msg );
				return;
			} else
			if ( msgType.equals( "3" )) {
				rejected( msg );
			} else
			if ( msgType.equals( "j" )) {
				businessRejected( msg );
			}
			fireMessageReceived( ev );
		} catch ( Exception ex ) {
			logger.error( this, "error", ex );
		}
	}

	synchronized void resend( FIXMessage req ) throws FIXException {
		int begin = 0;
		int end = 0;
		try {
			begin = Integer.valueOf( req.getField( "BeginSeqNo" )).intValue();
			end = Integer.valueOf( req.getField( "EndSeqNo" )).intValue();
		} catch ( Exception ex ) {
			throw( new FIXException( "illegal sequence no" ));
		}
		logger.debug( this, "resend " + begin + "-" + end );
		if ( begin > end ) {
			return;
		}

		boolean result = true;
		for ( int i = begin; i <= end; i++ ) {
			FIXMessage msg = null;
			try {
				store.load( this, i );
			} catch ( Exception ex ) {
				logger.error( this, "message load error " + i, ex );
			}
			// @@@
			// 見つからなかったとき
			if ( msg == null ) {
				// sendDummy( req, i );
				result = false;
				break;
			}
			logger.debug( this, "resend " + i );
			socket.send( msg );
		}

		if ( !result ) {
			resetSequence( begin, end );
		}
	}

	void sendDummy( FIXMessage req, int seq ) throws FIXException {
		FIXMessage msg = getParser().createMessage( "0" );
		SimpleDateFormat sdf = new SimpleDateFormat( "yyyyMMdd-HH:mm:ss" );
		sdf.setTimeZone( TimeZone.getTimeZone( "UTC" ));
		Date now = new Date();
		msg.setField( "SendingTime", sdf.format( now ));
		msg.setField( "MsgSeqNum", String.valueOf( seq ));
		msg.setField( "BeginString", version );
		msg.setField( "SenderCompID", senderCompID );
		msg.setField( "TargetCompID", targetCompID );
	}

	void resetSequence( int begin, int end ) throws FIXException {
		logger.debug( this, "resetSequence " + begin + "-" + end );
		FIXMessage msg = getParser().createMessage( "4" );
		msg.setField( "NewSeqNo", String.valueOf( end + 1 ));
		send( msg );
	}

	void rejected( FIXMessage msg ) throws FIXException {
		String seq = msg.getField( "RefSeqNum" );
		String reason = msg.getField( "SessionRejectReason" );
		String text = msg.getField( "Text" );
		logger.error( this, "rejected." + " RefSeqNum=" + seq );

		RejectedException ex = new RejectedException( seq, reason, text );
		fireExceptionOccurred( ex );
	}

	void businessRejected( FIXMessage msg ) throws FIXException {
		String seq = msg.getField( "RefSeqNum" );
		String reason = msg.getField( "BusinessRejectReason" );
		String text = msg.getField( "Text" );
		logger.error( this, "rejected." + " RefSeqNum=" + seq );

		RejectedException ex = new RejectedException( seq, reason, text );
		fireExceptionOccurred( ex );
	}

	void prepareMessage( FIXMessage msg ) throws FIXException {
		SimpleDateFormat sdf = new SimpleDateFormat( "yyyyMMdd-HH:mm:ss" );
		sdf.setTimeZone( TimeZone.getTimeZone( "UTC" ));
		Date now = new Date();
		msg.setField( "SendingTime", sdf.format( now ));
		int seq = store.getSendSeqNum() + 1;
		msg.setField( "MsgSeqNum", String.valueOf( seq ));
		if ( msg.getField( "BeginString" ) == null ) {
			if ( transportProtocol != null ) {
				msg.setField( "BeginString", transportProtocol );
			} else {
				msg.setField( "BeginString", version );
			}
		}
		msg.setField( "SenderCompID", senderCompID );
		msg.setField( "TargetCompID", targetCompID );
	}

	public synchronized void send( FIXMessage msg ) throws FIXException {
		SessionState state = getState();
		if ( state != SessionState.LOGONED &&
			!msg.getMsgType().equals( "A" ))
		{
			throw( new IllegalStateException( "not LOGONED." ));
		}

		prepareMessage( msg );
		socket.send( msg );
		int seq = Integer.valueOf( msg.getField( "MsgSeqNum" )).intValue();
		store.setSendSeqNum( seq );
		try {
			store.save( this, msg );
		} catch ( Exception ex ) {
			logger.error( this, "store error", ex );
		}
		hbChecker.resetLastSent();
		log( IMessageLogger.Direction.SEND, msg );
	}

	public void log( IMessageLogger.Direction dir, FIXMessage msg )
		throws FIXException
	{
		for ( IMessageLogger logger : vecLogger ) {
			if ( logger.getDirection() == IMessageLogger.Direction.BOTH ||
				logger.getDirection() == dir )
			{
				logger.log( dir, msg );
			}
		}
	}

	public void reject( FIXMessage req, SessionRejectReason reason )
		throws FIXException
	{
		FIXMessage msg = getParser().createMessage( "3" );
		msg.setField( "RefSeqNum", req.getField( "MsgSeqNum" ));
		msg.setField( "RefMsgType", req.getField( "MsgType" ));
		msg.setField( "SessionRejectReason", String.valueOf( reason.getCode()) );
		send( msg );
	}

	public void setLogger( List<IMessageLogger> var ) {
		vecLogger.clear();
		vecLogger.addAll( var );
	}

	public String toString() {
		String ret = "FIXSession id=" + getSessionId();
		ret += " heartbeat=" + getHeartBeat()
			+ " heartbeatTimeout=" + getHeartBeatTimeout();
		ret += " socket=[" + getSocket() + "]";
		return ret;
	}

	String	sessionId;	// セッションID
	FIXSocket	socket;	// ソケット
	IMessageParser	parser;	// メッセージ解析クラス
	int	timeout;	// メッセージ受信タイムアウト
	boolean	debug;	// debugフラグ
	String	version;	// FIXのバージョン
	String	transportProtocol;	// FIXのトランスポートレイヤのバージョン
	long	heartbeat;	// HeartBeat間隔
	long	heartbeatTimeout;	// HeartBeatタイムアウト時間
	String	senderCompID;	// 送信元ID
	String	targetCompID;	// 相手先ID

	/**
	* セッションIDを設定します。
	* @param	var		設定するデータ
	*/
	public void setSessionId( String var ) {
		this.sessionId = var;
	}
	/**
	* セッションIDを取得します。
	* @return	セッションID
	*/
	public String getSessionId() {
		return sessionId;
	}

	/**
	* ソケットを設定します。
	* @param	var		設定するデータ
	*/
	public void setSocket( FIXSocket var ) {
		this.socket = var;
	}
	/**
	* ソケットを取得します。
	* @return	ソケット
	*/
	public FIXSocket getSocket() {
		return socket;
	}

	/**
	* メッセージ解析クラスを設定します。
	* @param	var		設定するデータ
	*/
	public void setParser( IMessageParser var ) {
		this.parser = var;
	}
	/**
	* メッセージ解析クラスを取得します。
	* @return	メッセージ解析クラス
	*/
	public IMessageParser getParser() {
		return parser;
	}

	/**
	* メッセージ受信タイムアウトを設定します。
	* @param	var		設定するデータ
	*/
	public void setTimeout( int var ) {
		this.timeout = var;
	}
	/**
	* メッセージ受信タイムアウトを取得します。
	* @return	メッセージ受信タイムアウト
	*/
	public int getTimeout() {
		return timeout;
	}

	/**
	* debugフラグを設定します。
	* @param	var		設定するデータ
	*/
	public void setDebug( boolean var ) {
		this.debug = var;
	}
	/**
	* debugフラグを取得します。
	* @return	debugフラグ
	*/
	public boolean isDebug() {
		return debug;
	}

	/**
	* FIXのバージョンを設定します。
	* @param	var		設定するデータ
	*/
	public void setVersion( String var ) {
		this.version = var;
	}
	/**
	* FIXのバージョンを取得します。
	* @return	FIXのバージョン
	*/
	public String getVersion() {
		return version;
	}

	/**
	* FIXのトランスポートレイヤのバージョンを設定します。
	* @param	var		設定するデータ
	*/
	public void setTransportProtocol( String var ) {
		this.transportProtocol = var;
	}
	/**
	* FIXのトランスポートレイヤのバージョンを取得します。
	* @return	FIXのトランスポートレイヤのバージョン
	*/
	public String getTransportProtocol() {
		return transportProtocol;
	}

	/**
	* HeartBeat間隔を設定します。
	* @param	var		設定するデータ
	*/
	public void setHeartBeat( long var ) {
		this.heartbeat = var;
	}
	/**
	* HeartBeat間隔を取得します。
	* @return	HeartBeat間隔
	*/
	public long getHeartBeat() {
		return heartbeat;
	}

	/**
	* HeartBeatタイムアウトを設定します。
	* @param	var		設定するデータ
	*/
	public void setHeartBeatTimeout( long var ) {
		this.heartbeatTimeout = var;
	}
	/**
	* HeartBeatタイムアウトを取得します。
	* @return	HeartBeatタイムアウト
	*/
	public long getHeartBeatTimeout() {
		return heartbeatTimeout;
	}

	/**
	* 送信元IDを設定します。
	* @param	var		設定するデータ
	*/
	public void setSenderCompID( String var ) {
		this.senderCompID = var;
	}
	/**
	* 送信元IDを取得します。
	* @return	送信元ID
	*/
	public String getSenderCompID() {
		return senderCompID;
	}

	/**
	* 相手先IDを設定します。
	* @param	var		設定するデータ
	*/
	public void setTargetCompID( String var ) {
		this.targetCompID = var;
	}
	/**
	* 相手先IDを取得します。
	* @return	相手先ID
	*/
	public String getTargetCompID() {
		return targetCompID;
	}


}
