package jp.co.sds_corp.fixconnect.transport;

import java.util.Date;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import jp.co.sds_corp.fixconnect.FIXException;
import jp.co.sds_corp.fixconnect.message.FIXMessage;
import jp.co.sds_corp.fixconnect.message.IMessageReceiver;
import jp.co.sds_corp.fixconnect.message.MessageReceiveEvent;
import jp.co.sds_corp.util.config.DOMConfiguration;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class StressRunner
	implements IMessageReceiver, IStateChangeListener, ITransportEventListener
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	Initiator	initiator;
	List<FIXMessage> listMsg = new Vector<FIXMessage>();
	long	interval = 1000;
	long	delay = 1000;
	int		times = 100;
	int		count;
	int		report = 1000;

	public void setUp() throws Exception {
		DOMConfiguration doc = DOMConfiguration.findResource( "stress.xml" );

		IConfiguration config = doc.getChild( "Client" );
		setUpInitiator( config );
		setUpStress( config );

		createMessageList( config );
		logger.info( this, "send " + listMsg.size() + " messages." );
	}

	void setUpInitiator( IConfiguration config ) throws Exception {
		initiator = new Initiator();
		initiator.init( config );

		initiator.addMessageReceiver( this );
		initiator.addEventListener( this );
		initiator.addStateChangeListener( this );

		initiator.start();
    }

	void setUpStress( IConfiguration config ) throws Exception {
		IConfiguration attr = config.getAttributes();

		String var;
		var = attr.get( "interval", null );
		if ( var != null && var.length() > 0 ) {
			interval = Long.valueOf( var ).longValue();
		}
		var = attr.get( "delay", null );
		if ( var != null && var.length() > 0 ) {
			delay = Long.valueOf( var ).longValue();
		}
		var = attr.get( "times", null );
		if ( var != null && var.length() > 0 ) {
			times = Integer.valueOf( var ).intValue();
		}
		var = attr.get( "report", null );
		if ( var != null && var.length() > 0 ) {
			report = Integer.valueOf( var ).intValue();
		}
	}

	void createMessageList( IConfiguration config ) throws Exception {
		List<IConfiguration> list = config.listChild( "message" );
		for ( IConfiguration element : list ) {
			FIXMessage message = createMessage( element );
			listMsg.add( message );
		}
	}

	FIXMessage createMessage( IConfiguration config ) throws Exception {
		FIXMessage msg;
		FIXSession session = initiator.getSession();

		String msgType = config.getAttributes().get( "MsgType" );
		msg = session.getParser().createMessage( msgType );
		List<IConfiguration> list = config.listChild( "field" );
		for ( IConfiguration element : list ) {
			String name = element.getAttributes().get( "name" );
			String value = element.getAttributes().get( "value" );
			msg.setField( name, value );
		}
		return msg;
	}

	public void onMessage( MessageReceiveEvent ev ) {
		FIXMessage msg = ev.getMessage();
		// logger.debug( this, "message: " + msg );
	}

	public synchronized void onEvent( TransportEvent ev ) {
		logger.debug( this, "event:" + ev );
	}
	public void onException( FIXException ex ) {
		logger.debug( this, "exception:" + ex, ex );
	}

	public void stateChanged( StateChangeEvent ev ) {
		logger.debug( this, "state changed:" + ev );
	}

	Object lock = new Object();
    /**
     */
    public void testSession() throws Exception {
		if ( listMsg.size() == 0 ) {
			throw( new Exception( "no messages defined." ));
		}
		logger.debug( this, "send " + listMsg.size() + " messages." );
		for ( FIXMessage message : listMsg ) {
			logger.debug( this, "message:\n" + message );
//			logger.telegram( "message:", message.getBytes(), false );
		}

		logon();

		StressTask task = new StressTask();
		Date start = new Date();
		Timer timer = new Timer();
		timer.scheduleAtFixedRate( task, delay, interval );

		for ( ;; ) {
			synchronized ( lock ) {
				lock.wait();
			}
			if ( getCount() > times ) {
				break;
			}
		}
		timer.cancel();

		Date end = new Date();
		logger.debug( this, "###########################" );
		logger.debug( this, "elapsed time=" + ( end.getTime() - start.getTime() ));
		logger.debug( this, "###########################" );

		logout();
	}

	class StressTask extends TimerTask {
		public void run() {
			try {
				int count = getCount();
				if ( report > 0 && ( count % report == 0 )) {
					logger.info( this, "send " + count + " messages." );
				}
				FIXMessage message = listMsg.get( count % listMsg.size() );
				FIXSession session = initiator.getSession();
				session.send( message );

				setCount( count + 1 );
				if ( getCount() > times ) {
					cancel();
					synchronized ( lock ) {
						lock.notifyAll();
					}
					return;
				}
			} catch ( Exception ex ) {
				logger.error( this, "send error", ex );
			}
		}
	}

	synchronized int getCount() {
		return count;
	}
	synchronized void setCount( int var ) {
		this.count = var;
	}

    /**
     */
    public void logon() throws Exception {
		initiator.logon();
	}

    /**
     */
    public void logout() throws Exception {
		initiator.logout();
    }

    public void tearDown() throws Exception {
		logger.debug( this, "tearDown" );
		initiator.stop();
		initiator.addMessageReceiver( this );
		initiator.addEventListener( this );
		initiator.addStateChangeListener( this );
    }

	public static void main( String[] args ) {
		StressRunner obj = new StressRunner();
		try {
			obj.setUp();
			obj.testSession();
			obj.tearDown();
		} catch ( Exception ex ) {
			logger.error( StressRunner.class, "runner error", ex );
		}
	}

}
